"""
Logging support for Ten8t.

The purpose of this module is to setup package logging in a general way that integrates
with other packages and loggers.

The key feature is that you can call ten8t_setup_logging AFTER you have configured
other packages' loggers' and it will just work along with them.

If the only logging in the system is setup with this tool then you can still
use this and provide a target stream (e.g., stdout) and a file name.

This method is set up to manage a single global logger object named ten8t_logger
that should be used by all the ten8t modules.

"""
import logging

# Create a library-specific logger
ten8t_logger = logging.getLogger("ten8t")
ten8t_logger.addHandler(logging.NullHandler())  # Default to NullHandler


def ten8t_setup_logging(
        level: int = logging.WARNING,
        propagate: bool = True,
        file_name: str | None = None,
        stream_=None,
        format_string: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> None:
    """
    Configure the global logger for the 'ten8t' package.

    Args:
        level (int): Logging level (default=logging.WARNING).
        propagate (bool): If True, propagates logs to the parent logger (default=True).
        file_name (str): Optional file path to save logs.
        stream_ (io.TextIOWrapper | None): Optional stream (e.g., sys.stdout, sys.stderr).
        format_string (str): Format string for log messages (default includes timestamp, name, level, and message).

    Returns:
        None: Configures the global `ten8t_logger`.
    """
    global ten8t_logger

    # Create a formatter
    formatter = logging.Formatter(format_string)

    # Add a file handler if file_name is provided
    if file_name:
        try:
            file_handler = logging.FileHandler(file_name)
            file_handler.setFormatter(formatter)
            ten8t_logger.addHandler(file_handler)
        except (OSError, PermissionError) as e:
            raise ValueError(f"Error setting up file handler for '{file_name}'. Details: {e}") from e

    # Add a stream handler if stream is specified
    if stream_:
        if not hasattr(stream_, "write"):
            raise ValueError(f"The provided stream '{stream_}' is not a valid writable stream.")
        stream_handler = logging.StreamHandler(stream_)
        stream_handler.setFormatter(formatter)
        ten8t_logger.addHandler(stream_handler)

    # Configure logger settings
    ten8t_logger.setLevel(level)
    ten8t_logger.propagate = propagate


def ten8t_reset_logging():
    """
    Reset the logger to its initial state by removing all handlers and reinstating a NullHandler.

    NOTE: This function is intended for testing purposes ONLY.
          Using this in production code may disrupt logging configurations.
    """
    global ten8t_logger
    # Remove all existing handlers
    for handler in list(ten8t_logger.handlers):  # Make a copy to avoid modifying the list during iteration
        ten8t_logger.removeHandler(handler)

    # Add a NullHandler to restore the default silent state
    ten8t_logger.addHandler(logging.NullHandler())

    # Reset the logger's level and propagation settings
    ten8t_logger.setLevel(logging.NOTSET)
    ten8t_logger.propagate = False



"""
This example shows a NON-TRIVIAL example of how to use ten8t with streamlit.  An attempt is made
to use most of the major feature groups including:

1) Packages/Modules/Functions
2) Use of all attributes for classifying rule functions.
3) Use of attributes in non-trivial way for filtering
4) Use of "real" progress bar
5) Display of detailed results
6) Display of "scoreboard"
7) Display of the json
8) Rule functions returning various types of values.

"""

import streamlit as st

import ten8t as t8

st.set_page_config(layout='wide')


def color(c, t):
    """Streamlit color format"""
    return f':{c}[{t}]'


def green(text):
    """Streamlit green formatter."""
    return color('green', text)


def orange(text):
    """Streamlit orange formatter. """
    return color('orange', text)


def blue(text):
    """Streamlit blue formatter. """
    return color('blue', text)


def red(text):
    """Streamlit red formatter. """
    return color('red', text)


def violet(text):
    """Streamlit violet formatter."""
    return color('violet', text)


def display_overview(checker: t8.Ten8tChecker) -> None:
    """
    Display the results from the checker in a scoreboard format.

    Args:
        checker: Ten8t checker
    """
    table_headers = "| Metric | Value |"
    table_separators = "|:----:|:----:|"

    # Define the data rows
    table_rows = [
        f"| {green('**SCORE**')} | {checker.score:0.1f}% |",
        f"| {green('**PASS**')} | {checker.pass_count} |",
        f"| {red('**FAIL**')} | {checker.fail_count} |",
        f"| {orange('**WARN**')} | {checker.warn_count} |",
        f"| {violet('**SKIP**')} | {checker.skip_count} |",
        f"| {blue('**TOTAL**')} | {checker.result_count} |",
    ]

    # Join headers, separators, and rows to form a complete Markdown table
    markdown_table = "\n".join([table_headers, table_separators, *table_rows])

    # Display the Markdown table in Streamlit
    st.markdown(markdown_table)


def yes_or_none(var):
    """Syntactic sugar."""
    return 'Yes' if bool(var) else ''


def display_results(results: list[t8.Ten8tResult]):
    """
    Display the results in a Markdown table

    Args:
        results: list of ten8t results
    """
    headers = ['Count', 'Status', 'Warn', 'Skipped', 'Tag', 'Level', 'Phase', 'RUID', 'Module Name', 'Function Name',
               'Message']

    # Start with the table headers and separators created as f-strings
    table = [
        f"| {' | '.join(headers)} |",
        f"|{' :----: |' * len(headers)}"
    ]

    for count, r in enumerate(results, start=1):
        # Determine the status
        status = "Skipped" if r.skipped else "Pass" if r.status else "Fail"

        c_f = green if r.status else red if not r.skipped else orange

        # Append each row directly to the table, note that I'm using the "rendered" view
        table.append(
            f"| {count} | {c_f(r.status)} | {orange(yes_or_none(r.warn_msg))} | {violet(yes_or_none(r.skipped))}| " \
            f"{blue(r.tag)} | {r.level} | {r.phase} | {r.ruid} | {c_f(r.module_name)} | {c_f(r.func_name)} " \
            f"|{c_f(r.msg_rendered)} |")

        # Convert the list of rows into a single string with line breaks
    markdown_table = "\n".join(table)

    st.markdown(markdown_table)


def display_package_info(pkg, checker):
    """
    Displays information about the checked package and available options for checking.

    Args:
        pkg: The package being checked.
        checker: The checker setup for validating the package.
    """

    st.title('Ten8t Demo')

    # Define the table headers
    table_headers = "| **Item** | **Value** |"
    table_separators = "|:----:|:----:|"

    # Define the data rows
    table_rows = [
        f"| **Package** | {pkg.name} |",
        f"| **Module Count** | {pkg.module_count} |",
        f"| **Function Count** | {checker.function_count} |",
        f"| **Tags** | {','.join(checker.tags)} |",
        f"| **Rule IDs** | {','.join(checker.ruids)} |",
        f"| **Levels** | {checker.levels} |",
        f"| **Phases** | {','.join(checker.phases)} |",
    ]

    # Combine all parts to form the full table
    markdown_table = "\n".join([table_headers, table_separators] + table_rows)

    # Display the table
    st.markdown(markdown_table)


def display_json_results(checker: t8.Ten8tChecker):
    """
    Display the JSON results of the given `t8.Ten8tChecker` object.

    Args:
        checker: An instance of `t8.Ten8tChecker` containing the JSON results to be displayed.
    """
    st.write("JSON Results")
    with st.expander("Results:", expanded=True):
        st.json(checker.as_dict())



class Ten8tStreamlitProgressBar(t8.Ten8tProgress):
    """ Implementation of a progress bar for streamlit. """

    def __init__(self, progress_bar: st.progress):
        """ Store streamlit progress bar in global state"""
        self.progress_bar = progress_bar

    def __call__(self, current_count: int, max_count: int, msg: str, result=None):
        """
        Display a status message/progress update.  The progress bar goes from 0 to full
        scale as a percentage of the number of function that have run, that is why
        we get current/max.  We also have a message which can be anything as well
        as a result which is the detailed results.  Presumably some progress systems
        would watch the result metadata while others would be content with the message.
        """
        max_count = max(max_count, 1)
        current_count = min(current_count, max_count)
        percent = min(max(0, current_count / float(max_count)), 1.0)
        self.progress_bar.progress(percent, msg)


def main():
    """
    Main method for running the Ten8t package checker.

    Usage:
        Run this method to run the Ten8t package checker. It prompts the user to select options for including or
        excluding functions based on tags, rule IDs, levels, and phases. After selecting the options,
        click the "Run Ten8t" button to start the rule checking process. The results will be displayed in the UI.

    Example:
        main()
    """

    packages_mapping = {
        "File System Checking": "../examples/file_system",
        "Generic": "../examples/my_package"
    }
    info_mkdown = """
    # Ten8t (pronounced ten-eighty) Checker Demonstrator
    
    This demonstrates the way the `ten8t` package can work by having sample checks residing in a folder and
    allowing those checks to be easily run with the json output displayed in a basic UI.
    """

    st.markdown(info_mkdown)
    package_name = st.selectbox("Select Package", options=list(packages_mapping.keys()), index=0)
    package_folder = packages_mapping[package_name]
    st_renderer = t8.Ten8tBasicStreamlitRenderer()
    package = t8.Ten8tPackage(folder=package_folder)
    checker = t8.Ten8tChecker(packages=[package], renderer=st_renderer, auto_setup=True)

    with st.container(border=True):
        with st.container(border=True):
            display_package_info(package, checker)
        include_ui = st.checkbox("Select Functions to Run By Including Items", value=True)

        if include_ui:
            st.write(
                "All of these options are ANDed together, if you select everything " \
                "from 1 of the lists all functions will be run.")
            tags = st.multiselect("Include These Tags", options=checker.tags, default=checker.tags)
            ruids = st.multiselect("Include These Rule Ids", options=checker.ruids, default=[])
            levels = st.multiselect('Include These Levels', options=checker.levels, default=[])
            phases = st.multiselect('Include These Phases', options=checker.phases, default=[])
        else:
            st.write(
                "All of these options are ANDed together, if you select everything from 1 of " \
                "the lists no functions will be run.")
            tags = st.multiselect("Exclude These Tags", options=checker.tags, default=None)
            ruids = st.multiselect("Exclude These Rule Ids", options=checker.ruids, default=None)
            levels = st.multiselect('Exclude These Levels', options=checker.levels, default=None)
            phases = st.multiselect('Exclude These Phases', options=checker.phases, default=None)

    if st.button("Run Ten8t"):
        if include_ui:
            checker.include_by_attribute(tags=tags, ruids=ruids, levels=levels, phases=phases)
        else:
            checker.exclude_by_attribute(tags=tags, ruids=ruids, levels=levels, phases=phases)
        if len(checker.collected) == 0:
            st.warning("No rule functions found. It appears that you have filtered everything out.")
        else:
            prog_bar = st.progress(0, text=f"Rule checking status 0 of {checker.function_count}")

            # Install progress bar that is nice for ten8t.
            checker.progress_callback = Ten8tStreamlitProgressBar(prog_bar)

            # Magic happens here
            results: list[t8.Ten8tResult] = checker.run_all()

            with st.container(border=True):
                display_overview(checker)

            with st.container(border=True):
                display_results(results)

            with st.container(border=True):
                display_json_results(checker)

            with st.container(border=True):
                display_full_results(checker)

if __name__ == "__main__":
    main()



"""Sample Ten8t module with check functions"""

import time

from ten8t import TR, attributes


@attributes(tag="tag1", ruid="m1_f1", level=1, phase="proto")
def check_module1_func1():
    """ Simple always passing function"""
    time.sleep(.5)
    yield TR(status=True, msg="Always passes")


@attributes(tag="tag2", ruid="m1_f2", level=2, phase="proto")
def check_module1_func2():
    """ Simple always Failing Function"""
    time.sleep(.5)
    yield TR(status=True, msg="Always passes")


@attributes(tag="tag3", ruid="m1_f3", level=3, phase="production")
def check_module1_func3():
    """ Skips because flag set"""
    time.sleep(.5)
    yield TR(status=False, msg="Always passes", skipped=True)



"""Sample ten8t module with check functions"""

import time

from ten8t import TR, attributes


@attributes(tag="tag1", ruid="m2_f1", level=1, phase="proto")
def check_module2_func1():
    "Another always passing function"
    time.sleep(.5)
    yield TR(status=True, msg="Always passes")


@attributes(tag="tag2", ruid='m2_f2', level=2, phase="production")
def check_module2_func2():
    "THis thing always fails"
    time.sleep(.5)
    yield TR(status=False, msg="Always fails")


@attributes(tag="tag2", ruid='m2_f3', level=2, phase="production")
def check_module2_func3():
    "This thing always warns"
    time.sleep(.5)
    yield TR(status=True, warn_msg="Always warns and passes")


@attributes(tag="tag2", ruid='m2_f4', level=2, phase="production")
def check_module2_func4():
    "This thing always warns"
    time.sleep(.5)
    yield TR(status=False, warn_msg="Always warns and fails")



"""Sample ten8t module with check functions"""
from ten8t import attributes, rule_files


@attributes(tag="folder", ruid="f1", level=1, phase="proto")
def check_folder1():
    """ Simple always passing function"""
    yield from rule_files.rule_path_exists('../examples/file_system/folder1')


@attributes(tag="folder", ruid="f2", level=1, phase="proto")
def check_folder2():
    """ Simple always passing function"""
    yield from rule_files.rule_path_exists('../examples/file_system/folder2')


@attributes(tag="folder", ruid="file2", level=1, phase="proto")
def check_files_f2():
    """ Simple always passing function"""
    for f in ['file1.txt', 'file2.txt']:
        yield from rule_files.rule_path_exists(f'../examples/file_system/folder2/{f}')


@attributes(tag="folder", ruid="file1", level=1, phase="proto")
def check_files_f1():
    """ Simple always passing function"""
    for f in ['file1.txt', 'file2.txt']:
        yield from rule_files.rule_path_exists(f'../examples/file_system/folder1/{f}')



"""Trivial Typer App to run Ten8t checks on a given target."""

import json
import logging
import pathlib
import sys

import typer
import uvicorn

import ten8t as t8
import ten8t_api

s = pathlib.Path('./src').resolve()
sys.path.insert(0, str(s))

# Declare at top of file so  decorators work as expeted.
app = typer.Typer(add_completion=False)


def dump_results(results):
    """Dump results to stdout"""
    for result in results:
        typer.echo(result)


def pretty_print_json(json_obj):
    """
    Pretty print a JSON object, converting non-string values to strings.

    Args:
        json_obj (dict): The JSON object to be pretty printed.

    Returns:
        str: The pretty printed JSON string.
    """

    def convert_non_strings(obj):
        if isinstance(obj, dict):
            return {k: convert_non_strings(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [convert_non_strings(elem) for elem in obj]
        else:
            return str(obj)

    pretty_json = json.dumps(convert_non_strings(json_obj), indent=4)
    return pretty_json


@app.command()
def run_checks(
        module: str = typer.Option(None, '-m', '--mod', help='The module to run rules against.'),
        pkg: str = typer.Option(None, '--pkg', help='The package to run rules against.'),
        json_file: str = typer.Option(None, '-j', '--json', help='The JSON file to write results to.'),
        flat: bool = typer.Option(False, '-f', '--flat', help='Should the output be flat or a hierarchy.'),
        score: bool = typer.Option(False, '-s', '--score', help='Print the score of the rules.'),
        api: bool = typer.Option(False, '-a', '--api', help='Start FastAPI.'),
        port: int = typer.Option(8000, '-p', '--port', help='FastAPI Port'),
        verbose: bool = typer.Option(False, '-v', '--verbose', help='Enable verbose output.'),
):
    """Run Ten8t checks on a given package or module from command line."""

    if verbose:
        t8.ten8t_setup_logging(level=logging.DEBUG, file_name="ten8t_cli.log")
    else:
        t8.ten8t_setup_logging(level=logging.INFO, file_name="ten8t_cli.log")

    t8.ten8t_logger.debug("Module=%s", module)
    t8.ten8t_logger.debug("pkg=%s", pkg)
    t8.ten8t_logger.debug("json_file=%s", json_file)
    t8.ten8t_logger.debug("flat=%s", flat)
    t8.ten8t_logger.debug("score=%s", score)
    t8.ten8t_logger.debug("api=%s", api)
    t8.ten8t_logger.debug("port=%s", port)
    t8.ten8t_logger.debug("verbose=%s", verbose)

    try:
        mod = None
        if module:
            target_path = pathlib.Path(module)
            if target_path.is_file():
                mod = t8.Ten8tModule(module_name=target_path.stem, module_file=str(target_path))
            else:
                typer.echo(f'Invalid module: {module} is not a file.')

        if pkg:
            folder = pathlib.Path(pkg)
            if folder.is_dir():
                pkg = t8.Ten8tPackage(folder=folder)
            else:
                typer.echo(f'Invalid package: {pkg} is not a folder.')

        # If they supply 1 or both they are all run since the checker can handle arbitrary combinations
        if mod or pkg:
            ch = t8.Ten8tChecker(modules=mod, packages=pkg, auto_setup=True)
            if api:
                ten8t_api.set_ten8t_checker(ch)
                uvicorn.run(ten8t_api.app, host='localhost', port=port)
                return
            else:
                t8.ten8t_logger.info("Running %s check functions.", ch.function_count)
                results = ch.run_all()
                t8.ten8t_logger.info("Pass=%s Fail=%s Skip=%s", ch.pass_count, ch.fail_count, ch.skip_count)
        else:
            typer.echo('Please provide a module, package to run checks on.')
            return

        if not results:
            typer.echo('There were no results.')
            return

        if not flat:
            # Show the results in a hierarchy rather than the flat version the data is stored as.
            results = t8.group_by(results, ['pkg_name', 'module_name', 'func_name'])
            pretty_data = pretty_print_json(results)
            typer.echo(pretty_data)
            return

        if verbose:
            dump_results(results)
        else:
            typer.echo(t8.overview(results))

        if score:
            test_score = t8.ScoreByResult()
            typer.echo(f'Score: {test_score(results):.1f}')

        if json_file:
            d = t8.ten8t_result.results_as_dict(results)
            with open(json_file, 'w', encoding='utf-8') as f:
                json.dump(d, f, indent=2)

    except t8.Ten8tException as e:
        typer.echo(f'Ten8tException: {e}')

    # Crude
    except Exception as e:
        typer.echo(f'An error occurred: {e}')


if __name__ == '__main__':
    app()
